/*
 * audina.c - part of Audina audiobook stretcher
 * Copyright (C) Bohdan R. Rau 2010-2011 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "audina.h"
#include <string.h>
#include <getopt.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>

static int detect_filetype(char *filename)
{
    char *c=strrchr(filename,'.');
    if (strstr(filename,"://")) return AUDINA_FMAT_AVC;
    if (!c) return AUDINA_FMAT_SNDFILE;
    if (!strcasecmp(c,".mp3")) return AUDINA_FMAT_MP3;
    if (!strcasecmp(c,".mpg") ||
        !strcasecmp(c,".mpeg") ||
        !strcasecmp(c,".awb") ||
        !strcasecmp(c,".mp4") ||
        !strcasecmp(c,".avi") ||
        !strncasecmp(c,".m4",3) ||
        !strncasecmp(c,".wm",3)) return AUDINA_FMAT_AVC;
    return AUDINA_FMAT_SNDFILE;
}

int InputConversion(char *filename,int format)
{
    if (format == AUDINA_FMAT_AUTO) {
        format=detect_filetype(filename);
    }
    if (special_flags & AUDINA_F_VERBOSE) {
        printf("Decoder selected: %s\n",
        (format == AUDINA_FMAT_MP3)?MP3_DECODER_NAME:
#ifdef USE_AVCODEC
        (format == AUDINA_FMAT_AVC)?"ffmpeg":
#endif
        "sndfile");
    }
    if (format == AUDINA_FMAT_SNDFILE) {
        return input_convert_sndfile(filename);
    }
    if (format == AUDINA_FMAT_MP3) {
        return input_convert_mpg123(filename);
    }
#ifdef USE_AVCODEC
    if (format == AUDINA_FMAT_AVC) {
        return input_convert_avcodec(filename);
    }
#endif
    errno=ENOSYS;
    return -1;
}

double tempo=1.0;
int use_feats=0;
int max_pause=0;
double pause_length=1.0;
int set_type;
double density=1.0;
double mad_gain=1.0;
double pitch=1.0;
int special_flags = 0;
double eqpole[3];
int eqband[2]={500,3000};
int use_equalizer;
double eqgain=1.0;
int output_samplefreq=16000;
int output_mp3_brate=32;
int use_contrast=0;
double contrast_level;
int output_freq_set=0;
int libavc_audio_idx=-1;
int64_t recording_time=0;
int64_t skip_time=0;
int part_time=0;
int64_t partial_samples;

int64_t sample_length;
int64_t current_samples;

char *title=NULL,*artist=NULL,*album=NULL;
int initial_track=0,current_track=0;

int input_is_raw=0;
int input_raw_channels;
int input_raw_type;
int input_raw_rate;
int output_padding=0;



static char *progname;
void help(void)
{
    fprintf(stderr,"Audina v %s - audiobook corrector\n",PACKAGE_VERSION);
    fprintf(stderr,"Usage is:\n%s [options] <infile> [<outfile>]\n",progname);
    fprintf(stderr,"Options are:\n");
    fprintf(stderr,"  --decoder=letter\n");
    if (isatty(2)) fprintf(stderr,
#ifdef USE_AVCODEC
        "  -i letter - input decoder (\x1b[4ma\x1b[0muto, \x1b[4ms\x1b[0mndfile, \x1b[4mm\x1b[0mpg123, \x1b[4mf\x1b[0mfmpeg)\n");
#else
        "  -i letter - input decoder (\x1b[4ma\x1b[0muto, \x1b[4ms\x1b[0mndfile, \x1b[4mm\x1b[0mpg123)\n");
#endif
    else fprintf(stderr,
#ifdef USE_AVCODEC
        "  -i letter - input decoder (auto, sndfile, mpg123, ffmpeg\n");
    fprintf(stderr,"\
              default is auto for files, ffmpeg for streams\n");
#else
        "  -i letter - input decoder (auto, sndfile, mpg123\n");
#endif
	fprintf(stderr,"\
  --raw-input=channels:type:rate\n\
  -I channels:type:rate - raw input data, where:\n\
     channels can be 1, 2, m or s\n\
     type can be s16, s24, s32 or f\n\
     Implies sndfile input decoder.\n\
  --speed=double\n\
  -s double - reading speed factor, 0.25 to 4.0\n\
  --pauses=double\n\
  -p double - pauses length factor, 0.25 to 4.0\n\
  --pitch=double\n\
  -P double - pitch as factor (0.5 to 2.0)\n\
              or in cents (-1200 to 0 and 2 to 1200)\n\
  --max-pause=int\n\
  -m int    - maximum pause (msec), 200 to 2000\n\
  --pad=int\n\
            - adds silence (in msec) at end of output\n\
  --density=double\n\
  -d double - silence detector density  (with optional %% or dB)\n\
              default 1.0\n\
  --gain=double\n\
  -g double -  input gain (with optional %% or dB), default 1.0\n\
  --output-type=type\n\
  -t type  - output file type (mp3, wav, awb or raw)\n\
  --equalizer=dbl:dbl:dbl\n\
  -E dbl:dbl:dbl - 3-band equalizer (dB, -20..20)\n\
  --equalizer-band=int:int\n\
  -B int:int     - equalizer low:high cutoff frequency\n\
                   low range: 50..1500, high range: 1800..8000\n\
                   default 500:3000\n\
  --equalizer-gain=double\n\
  -G double - equalizer gain (with optional %% or dB)\n\
              default 1.0/max(equalizer band gains)\n\
  --samplerate=int\n\
  -f int    - output samplerate in kHz (16,22,32,44 or 48)\n\
  --audio-index=int\n\
  -A int    - audio stream index (0..max index, default: first found)\n\
  --mp3-bitrate=int\n\
  -b int    - output mp3 bitrate in kbps (32,48 or 64)\n\
  --contrast=int\n\
  -c int    - contrast effect (0 to 100 or 'none')\n\
  --from=h:m:s\n\
  -F h:m:s  - skip start (fails if impossible)\n\
  --record-time=h:m:s\n\
  -R h:m:s  - record time\n\
  --part-time=h:m:s\n\
  -r h:m:s  - part length\n\
  --script\n\
  -S        - script mode (does not catch CTRL-C)\n\
  --show-peak\n\
  -V        - show input peak amplitude\n\
  --verbose\n\
  -v        - print summary\n\
  --show-energy\n\
  -D        - show each frame energy\n\
  --id3-artist=string\n\
  --id3-album=string\n\
  --id3-title=string\n\
            - set id3 strings. If not set, these strings\n\
              will be copied from source file if available.\n\
  --id3-track=int\n\
            - initial track number (will be increased by 1\n\
              on each part of multipart output). If not set,\n\
              no track information will be stored.\n\
  --disable-id3\n\
            - turns off writing id3 tags\n\
  --tune\n\
            - tuning mode. Each pause will be exactly max-pause\n\
              milliseconds long.\n");
#ifdef HAVE_PULSE
fprintf(stderr,"  --pulse\n\
            - use pulseaudio driver to play\n");
#endif
fprintf(stderr,"  --help\n\
  -h        - this help\n\
If -t option is not set, outname must have one of\n\
extensions: mp3, wav or awb.\n\
If outfile name is omitted, program plays the output wave\n\
using default libao%s driver.\n\
In case of raw mode output samplerate must be set.\n\
Raw output is always 16-bit little-endian mono.\n\
Encoding to stdout is possible only in raw mode.\n\
Decoding from stdin is possible only in raw mode.\n\
Verbose mode and multipart cannot be used when writing to stdout.\n",
#ifdef HAVE_PULSE
    " or pulseaudio (if selected)"
#else
    ""
#endif
);
    exit(1);
}

void read_raw_input(char *opt)
{
	char c;
	input_is_raw=1;
	c=*opt++;
	if (c=='1' || c == 'm') input_raw_channels=1;
	else if (c=='2' || c == 's') input_raw_channels=2;
	else help();
	if (*opt++ != ':') help();
	if (!strncmp(opt,"s16:",4)) {
		input_raw_type=RAW_ITYPE_S16;opt+=4;
	}
	else if (!strncmp(opt,"s24:",4)) {
		input_raw_type=RAW_ITYPE_S24;opt+=4;
	}
	else if (!strncmp(opt,"s32:",4)) {
		input_raw_type=RAW_ITYPE_S32;opt+=4;
	}
	else if (!strncmp(opt,"f:",2)) {
		input_raw_type=RAW_ITYPE_FLOAT;opt+=2;
	}
	else help();
	if (!*opt || !isdigit(*opt)) help();
	input_raw_rate=strtol(opt,&opt,10);
	if (*opt || input_raw_rate < 1) help();
}

double my_strtod(char *inp,char **outp)
{
    char buf[strlen(inp)+1],*eob,*c;
    double rc;
    strcpy(buf,inp);
    for (c=buf;*c;c++) if (*c==',') *c='.';
    rc=strtod(buf,&eob);
    if (outp) *outp=inp+(eob-buf);
    return rc;
}

double gain(char *inp,char **outp)
{
    char *eob;
    double rc;
    rc=my_strtod(inp,&eob);
    while (*eob && isspace(*eob)) eob++;
    if (*eob=='%') {
        eob++;
        rc=rc/100.0;
    }
    else if (!strncasecmp(eob,"db",2)) {
        eob+=2;
        rc=pow(10.0,rc/20.0);
    }
    if (outp) *outp=eob;
    return rc;
}

int64_t hms(char *c)
{
    int64_t hx;
    hx=strtol(c,&c,10);
    if (*c++ == ':') hx=60 * hx + strtol(c,&c,10);
    if (*c++ == ':') hx=60 * hx + strtol(c,&c,10);
    return hx;
}

#define AUDINA_PARAM_ARTIST 501
#define AUDINA_PARAM_ALBUM 502
#define AUDINA_PARAM_TITLE 503
#define AUDINA_PARAM_TRACK 504
#define AUDINA_PARAM_NOID3 505
#define AUDINA_PARAM_TUNE 506
#define AUDINA_PARAM_ID3IE 507
#define AUDINA_PARAM_ID3OE 508
#define AUDINA_PARAM_MINPAUSE 509
#define AUDINA_PARAM_PAD 510
#define AUDINA_PARAM_PA 511


static struct option long_opts[]={
    {"help",0,NULL,'h'},
    {"speed",1,NULL,'s'},
    {"pauses",1,NULL,'p'},
    {"max-pause",1,NULL,'m'},
    {"output-type",1,NULL,'t'},
    {"decoder",1,NULL,'i'},
    {"raw-input",1,NULL,'I'},
    {"density",1,NULL,'d'},
    {"gain",1,NULL,'g'},
    {"script",0,NULL,'S'},
    {"show-peak",0,NULL,'V'},
    {"show-energy",0,NULL,'D'},
    {"verbose",0,NULL,'v'},
    {"pitch",1,NULL,'P'},
    {"equalizer",1,NULL,'E'},
    {"equalizer-band",1,NULL,'B'},
    {"equalizer-gain",1,NULL,'G'},
    {"contrast",1,NULL,'c'},
    {"mp3-bitrate",1,NULL,'b'},
    {"samplerate",1,NULL,'f'},
    {"audio-index",1,NULL,'A'},
    {"from",1,NULL,'F'},
    {"record-time",1,NULL,'R'},
    {"part-time",1,NULL,'r'},
    {"id3-album",1,NULL,AUDINA_PARAM_ALBUM},
    {"id3-artist",1,NULL,AUDINA_PARAM_ARTIST},
    {"id3-title",1,NULL,AUDINA_PARAM_TITLE},
    {"id3-track",1,NULL,AUDINA_PARAM_TRACK},
    {"disable-id3",0,NULL,AUDINA_PARAM_NOID3},
    {"id3-input-encoding",1,NULL,AUDINA_PARAM_ID3IE},
    {"id3-output-encoding",1,NULL,AUDINA_PARAM_ID3OE},
    {"tune",0,NULL,AUDINA_PARAM_TUNE},
    {"min-pause-frames",1,NULL,AUDINA_PARAM_MINPAUSE},
    {"pad",1,NULL,AUDINA_PARAM_PAD},
#ifdef HAVE_PULSE
    {"pulse",0,NULL,AUDINA_PARAM_PA},
#endif
    {NULL,0,NULL,0}};

#define fpause (special_flags & AUDINA_F_FORCEPAUSE)

int main(int argc,char *argv[])

{
    int fmat=AUDINA_FMAT_AUTO;
    FILE *verbf;
    progname=argv[0];
    for (;;) {
        int copt=getopt_long(argc,argv,
            "hs:p:m:t:d:g:P:E:G:DSVvc:b:f:i:B:A:R:r:F:I:",
            long_opts,NULL);
        if (copt < 0) break;
        if (copt == 'V') {
            special_flags |= AUDINA_F_VUMETER;
            continue;
        }
        if (copt == 'S') {
            special_flags |= AUDINA_F_SCRIPT;
            continue;
        }
        if (copt == AUDINA_PARAM_TUNE) {
            special_flags |= AUDINA_F_TUNE;
            continue;
        }
        if (copt == 'B') {
            int m,n;
            m=strtol(optarg,&optarg,10);
            if (*optarg++ != ':') help();
            n=strtol(optarg,&optarg,10);
            if (m<50 || m>1500) help();
            if (n<1800 || n>8000) help();
            eqband[0]=m;
            eqband[1]=n;
            continue;
        }
        if (copt == 'R') {
            recording_time=hms(optarg);
            continue;
        }
        if (copt == 'F') {
            skip_time=hms(optarg);
            continue;
        }
        if (copt == 'r') {
            part_time=hms(optarg);
            continue;
        }
        if (copt == 'A') {
            int n=strtol(optarg,NULL,10);
            if (n<0) help();
            libavc_audio_idx=n;
            continue;
        }
        if (copt == 'D') {
            special_flags |= AUDINA_F_DUMPF;
            use_feats |= AUDINA_FEAT_PAUSES;
            continue;
        }
        if (copt == 'i') {
            if (*optarg=='f') fmat=AUDINA_FMAT_AVC;
            else if (*optarg=='s') fmat=AUDINA_FMAT_SNDFILE;
            else if (*optarg=='m') fmat=AUDINA_FMAT_MP3;
            else if (*optarg=='a') fmat=AUDINA_FMAT_AUTO;
            else help();
            continue;
        }
        if (copt == 'c') {
            if (!strcasecmp(optarg,"none")) use_contrast=0;
            else {
                int n=strtol(optarg,NULL,10);
                if (n<0 || n> 100) help();
                use_contrast=1;
                contrast_level=0.03+n/1000.0;
            }
            continue;
        }
        if (copt == 'v') {
            special_flags |= AUDINA_F_VERBOSE;
            continue;
        }
        if (copt == 'f') {
            int n=strtol(optarg,NULL,10);
            output_freq_set=1;
            switch(n) {
                case 16: output_samplefreq=16000;break;
                case 22: output_samplefreq=22020;break;
                case 32: output_samplefreq=32000;break;
                case 44: output_samplefreq=44100;break;
                case 48: output_samplefreq=48000;break;
                default: help();
            }
            continue;
        }
        if (copt == 'b') {
            int n=strtol(optarg,NULL,10);
            if (n != 32 && n != 48 && n != 64) help();
            output_mp3_brate=n;
            continue;
        }
        if (copt == 't') {
            if (!strcmp(optarg,"mp3")) set_type=1;
            else if (!strcmp(optarg,"wav")) set_type=2;
            else if (!strcmp(optarg,"awb")) set_type=3;
            else if (!strcmp(optarg,"raw")) set_type=4;
            else help();
            continue;
        }
        if (copt == 's') {
            tempo=my_strtod(optarg,NULL);
            if (tempo<0.2 || tempo > 4.0) help();
            continue;
        }
        if (copt == 'P') {
            pitch=my_strtod(optarg,NULL);
            if (pitch<-1200.0 || pitch > 1200.0) help();
            if (pitch>0 && pitch < 0.5) help();
            if (pitch<0 || pitch > 2.001) pitch=pow(2.0,pitch/1200.0);
            continue;
        }
        if (copt=='p') {
            use_feats|=AUDINA_FEAT_PAUSES;
            pause_length=my_strtod(optarg,NULL);
            if (pause_length < 0.25 || pause_length > 4.0) help();
            continue;
        }
        if (copt=='m') {
            use_feats|=AUDINA_FEAT_MAXPAUSE;
            max_pause=strtol(optarg,NULL,10);
            if (max_pause < 50 || max_pause > 2000) help();
            continue;
        }
        if (copt== 'd') {
            density=gain(optarg,NULL);
            continue;
        }
        if (copt== 'g') {
            mad_gain=gain(optarg,NULL);
            continue;
        }
        if (copt == 'E') {
            double d;
            d=my_strtod(optarg,&optarg);
            if (errno || *optarg++ != ':' || d<-50 || d>20) help();
            eqpole[0]=pow(10.0,d/20.0);
            d=my_strtod(optarg,&optarg);
            if (errno || *optarg++ != ':' || d<-20 || d>20) help();
            eqpole[1]=pow(10.0,d/20.0);
            d=my_strtod(optarg,&optarg);
            if (errno || *optarg || d<-20 || d>20) help();
            eqpole[2]=pow(10.0,d/20.0);
            use_equalizer|=1;
            continue;
        }
        if (copt == 'G') {
            use_equalizer |= 2;
            eqgain=gain(optarg,NULL);
            continue;
        }
        if (copt == AUDINA_PARAM_ALBUM) {
            album=strdup(optarg);
            continue;
        }
        if (copt == AUDINA_PARAM_ARTIST) {
            artist=strdup(optarg);
            continue;
        }
        if (copt == AUDINA_PARAM_TITLE) {
            title=strdup(optarg);
            continue;
        }
        if (copt == AUDINA_PARAM_TRACK) {
            initial_track=current_track=strtol(optarg,NULL,10);
            continue;
        }
        if (copt == AUDINA_PARAM_NOID3) {
            special_flags |= AUDINA_F_NOID3;
            continue;
        }
        if (copt == AUDINA_PARAM_ID3IE) {
            id3_input_enc=strdup(optarg);
            continue;
        }
        if (copt == AUDINA_PARAM_ID3OE) {
            id3_output_enc=strdup(optarg);
            continue;
        }
        if (copt == AUDINA_PARAM_MINPAUSE) {
            min_pause_frames=strtol(optarg,NULL,10);
            if (min_pause_frames < 3 || min_pause_frames > 12) help();
            continue;
        }
        if (copt == AUDINA_PARAM_PAD) {
            output_padding=strtol(optarg,NULL,10);
            if (output_padding<0) help();
            continue;
        }
        if (copt == AUDINA_PARAM_PA) {
            use_pa=1;
            continue;
        }
        if (copt == 'I') {
			read_raw_input(optarg);
			continue;
		}
        help();
    }

    tempo=tempo/pitch;
    if (use_equalizer & 1) {
        if (!(use_equalizer & 2)) {
            eqgain=1.0/fmax(fmax(eqpole[0],eqpole[1]),eqpole[2]);
        }
        use_equalizer &= ~2;
    }
    if (part_time) {
        use_feats |= AUDINA_FEAT_PAUSES;
        special_flags |= AUDINA_F_FORCEPAUSE;
    }
    if (special_flags & AUDINA_F_TUNE) {
        special_flags |= AUDINA_F_FORCEPAUSE;
        use_feats |=AUDINA_FEAT_PAUSES;
        if (!(use_feats & AUDINA_FEAT_MAXPAUSE)) {
            max_pause=2000;
            use_feats |= AUDINA_FEAT_MAXPAUSE;
        }
    }
    if (use_feats & AUDINA_FEAT_PAUSES) {
        pause_length *= tempo;
        if (!(use_feats & AUDINA_FEAT_MAXPAUSE)) {
            if (!fpause && pause_length >= 0.95 && pause_length < 1.05) {
                pause_length=1.0;
                use_feats &= ~AUDINA_FEAT_PAUSES;
            }
            else{
                max_pause=2000;
                use_feats |= AUDINA_FEAT_MAXPAUSE;
            }
        }
    }
    if (use_feats & AUDINA_FEAT_MAXPAUSE) {
        max_pause *= tempo;
        use_feats |= AUDINA_FEAT_PAUSES;
    }
    if (tempo < 0.99 || tempo > 1.01) use_feats |= AUDINA_FEAT_TEMPO;
    if (optind != argc-2 && optind != argc-1) help();
    if (!strcmp(argv[optind],"-") && !input_is_raw) help();
    if (input_is_raw) fmat=AUDINA_FMAT_SNDFILE;
    if (set_type == 4 && ! output_freq_set) help();
    if (!set_type && optind == argc-2) {
        char *c=strrchr(argv[optind+1],'.');
        if (c) {
            if (!strcasecmp(c,".mp3")) set_type=1;
            else if (!strcasecmp(c,".wav")) set_type=2;
            else if (!strcasecmp(c,".awb")) set_type=3;
        }
        if (!set_type) help();
    }
    if (argv[optind+1] && !strcmp(argv[optind+1],"-")) {
        if(set_type != 4) help();
        if(part_time) help();
        if (special_flags & AUDINA_F_VERBOSE) help();
    }
    if (set_type == 3) output_samplefreq = 16000;
    if (special_flags & AUDINA_F_VERBOSE) {
        verbf=stdout;
        if (!isatty(1)) verbf=stderr;
        fprintf(verbf,"Equalizer: %s\n",(use_equalizer & 1)?"yes":"no");
        if (use_equalizer & 1) {
            fprintf(verbf,"Equalizer low gain: %5.3f\n",eqpole[0]);
            fprintf(verbf,"Equalizer low cut frequency: %d Hz\n",eqband[0]);
            fprintf(verbf,"Equalizer middle gain: %5.3f\n",eqpole[1]);
            fprintf(verbf,"Equalizer high gain: %5.3f\n",eqpole[2]);
            fprintf(verbf,"Equalizer high cut frequency: %d Hz\n",eqband[1]);
            fprintf(verbf,"Equalizer gain: %5.3f\n",eqgain);
        }
        if (use_feats & AUDINA_FEAT_TEMPO) fprintf(verbf,"Effective speed factor: %.3f\n",tempo);
        fprintf(verbf,"Effective pitch factor: %.3f\n",pitch);
        if (use_feats & AUDINA_FEAT_PAUSES) fprintf(verbf,"Effective pause factor: %.3f\n",pause_length);
        if (use_feats & AUDINA_FEAT_MAXPAUSE) fprintf(verbf,"Max pause: %d msec\n",max_pause);
        fprintf(verbf,"Input signal gain: %.3f\n",mad_gain);
        fprintf(verbf,"Silence treshold: %f\n",SILENCE_LIMIT/density);
        if (set_type == 2) fprintf(verbf,"MP3 bitrate: %d kbps\n",output_mp3_brate);
        if (set_type && recording_time) {
            fprintf(verbf,"Recording time: %ld seconds\n",(long)recording_time);
        }
        if (set_type && part_time) {
            fprintf(verbf,"Part length: %d seconds\n",part_time);
        }
    }
    if (encoder_init(argv[optind+1])) {
        if (errno) perror("Transcoder");
        exit(1);
    }
    if (InputConversion(argv[optind],fmat)) {
        if (errno) perror("Transcoder");
        exit(1);
    }
    return 0;
}

