/*
 * converter.c - part of Audina audiobook stretcher
 * Copyright (C) Bohdan R. Rau 2010-2011 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */


#include "audina.h"
#include <string.h>
#include <samplerate.h>
#include "sonic.h"
#include <math.h>
#include "equalizer.h"
#include <signal.h>

static void inline _simple_memcpy(register char *d1,register char *d2,int n)
{
    while (n-- > 0) *d1++ = *d2++;
}

#define simple_memcpy(a,b,c) _simple_memcpy((char *)(a),(char *)(b),c)

static SRC_STATE *resampler;
static SRC_DATA res_data;
static sonicStream sonic;
int samplefreq;

static int file_phase;
static int output_is_stdout;

/* equalizer */
struct EQSTATE eq_std,eq_dnl;

#define SILENT_PERSEC 32
#define PBUF_PAUSENO (2*SILENT_PERSEC)

int min_pause_frames=3;

static int frame_buffer_size,frame_buffer_pos;
static int pause_count;
static float *frame_buffer;
static float *pause_buffer,*pause_buffer_right,*second_buffer;
static int pause_buffer_frames;
static int max_pause_samples;
static int pause_left,pause_right;

int resampler_quality=SRC_SINC_MEDIUM_QUALITY;

static int force_finish;

void sig_handler(int dummy)
{
    force_finish=1;
}


int set_input_samplerate(int samplerate)
{
    int resampler_error;
    if (special_flags & AUDINA_F_VERBOSE) {
        long seconds=sample_length/samplerate;
        printf("Input samples %ld\n",(long)sample_length);
        printf("Input samplerate: %d\n",samplerate);
        printf("Input time %d:%02d:%02d\n",
            (int)(seconds/3600),
            (int)(seconds % 3600) / 60,
            (int)(seconds % 60));
        if (skip_time!=1) printf("Skip time: %d:%02d:%02d\n",
            (int)(skip_time/3600),
            (int)(skip_time % 3600) / 60,
            (int)(skip_time % 60));
    }
    samplerate *= pitch;
    if (special_flags & AUDINA_F_VERBOSE) {
        printf("Effective input samplerate: %d\n",samplerate);
    }
    if (samplerate < 6000) {
        fprintf(stderr,"Samplerate %d not supported\n",samplerate);
        errno=ENOSYS;
        return -1;
    }
    samplefreq=samplerate;
    if (!set_type && !output_freq_set) {
        int n,i;double d1,d2;
        int freqs[]={16000,22050,32000,44100,48000};
        for (i=0;i<5;i++) {
            d1=(double)freqs[i]/(double)samplefreq;
            if (d1<1.0) d1=1.0/d1;
            if (!i || d1<d2) n=i;
            d2=d1;
        }
        output_samplefreq=freqs[n];
    }
    if (special_flags & AUDINA_F_VERBOSE) printf("Output samplerate: %d\n",output_samplefreq);
    if (samplefreq != output_samplefreq) {
        resampler=src_new (resampler_quality, 1, &resampler_error) ;
        res_data.data_in=malloc(4096*sizeof(float));
        res_data.data_out=malloc(8200*sizeof(float));
        res_data.src_ratio=(double)output_samplefreq/(double)samplerate;
    }
    if (use_feats & 4) {
        sonic=sonicCreateStream(tempo,samplefreq);
    }
    if (use_equalizer & 1) {
        init_equaliser(&eq_std,eqband[0],eqband[1],samplefreq);
        eq_std.lg=eqpole[0];
        eq_std.mg=eqpole[1];
        eq_std.hg=eqpole[2];
    }
    file_phase=0;
    frame_buffer_size=samplefreq/SILENT_PERSEC;
    frame_buffer_pos=0;
    pause_count=0;
    if (use_feats) {
        frame_buffer=malloc(frame_buffer_size * sizeof(float));
        pause_buffer_frames=1+(max_pause*SILENT_PERSEC)/1000.0;
        if (pause_buffer_frames < 3) pause_buffer_frames=3;
        pause_buffer=malloc(frame_buffer_size * pause_buffer_frames *sizeof(float));
        second_buffer=malloc(frame_buffer_size * pause_buffer_frames *sizeof(float));
        pause_right=pause_buffer_frames/2;
        pause_left=pause_buffer_frames-pause_right;
        pause_buffer_right=pause_buffer+frame_buffer_size * pause_left;
        max_pause_samples=(max_pause * samplefreq)/1000.0;
    }
    if (!(special_flags & AUDINA_F_NOID3)) {
        if (special_flags & AUDINA_F_VERBOSE) {
            if (album) printf("ID3 Album: %s\n",album);
            if (artist) printf("ID3 Artist: %s\n",artist);
            if (title) printf("ID3 Title: %s\n",title);
            if (current_track) printf("ID3 Track: %d\n",current_track);
        }
    }
    return 0;
}

static int encoder_initialized;
static char *local_fname;
static long long coder_samples;
int coder_init(char *fname)
{
    int rc;
    static int handler_set=0;
    coder_samples=0;
    switch(set_type) {
        case 0: return init_ao_player();
        case 1: rc=init_lame_encoder(fname);break;
        case 2: rc=init_wave_encoder(fname,0);break;
        case 3: rc=init_amrwb_encoder(fname);break;
        case 4: rc=init_wave_encoder(fname,1);break;
        default:
            fprintf(stderr,"Internal error (unknown file type %d)\n",set_type);
            exit(1);
    }
    if (rc) return rc;
    if (!(special_flags & AUDINA_F_SCRIPT) && !handler_set) {
        signal(SIGINT,sig_handler);
        handler_set=1;
    }
    return 0;
}

int coder_encode(short *samples,int nsamp)
{
    coder_samples+=(long long)nsamp;
    partial_samples+=nsamp;
    if (recording_time && set_type && coder_samples/output_samplefreq >= recording_time) {
        force_finish=1;
    }
    switch(set_type) {
        case 0: return ao_play_samples(samples,nsamp);
        case 1: return lame_encode(samples,nsamp);
        case 4:
        case 2: return wave_encode(samples,nsamp);
        case 3: return encode_amrwb(samples,nsamp);
    }
    fprintf(stderr,"Internal error (unknown file type %d)\n",set_type);
    exit(1);
}

int coder_finish(void)
{
    long long coder_msec=coder_samples/(output_samplefreq/1000.0);
    fprintf((output_is_stdout?stderr:stdout),"Duration: %lld msec\n",coder_msec);
    switch(set_type) {
        case 0: return ao_stop();
        case 1: return lame_finish();
        case 4:
        case 2: return wave_finish();
        case 3: return finish_amrwb();
    }
    fprintf(stderr,"Internal error (unknown file type %d)\n",set_type);
    exit(1);
}

static double do_contrast(double d)
{
    if (d<-1.0) d=-1.0;
    else if (d>1.0) d=1.0;
    if (!use_contrast) return d;
    d=M_PI_2 * d;
    return sin(d+contrast_level * sin(d*4));
}

int encode_samples(float *samples,int nsamp)
{
    static short *bs,buffer[8200];int i,cnt,rc;
    if (!encoder_initialized) {
        rc=encoder_init(NULL);
        if (rc<0) return rc;
    }
    if (!samples) return encoder_finish();
    while (nsamp > 0) {
        cnt=nsamp;
        if (cnt > 4096) cnt=4096;
        bs=buffer;
        if (resampler) {
            for (i=0;i<cnt;i++) {
                res_data.data_in[i]=samples[i];
            }
            res_data.input_frames=i;
            res_data.output_frames=8200;
            res_data.end_of_input=0;
            src_process(resampler,&res_data);
            cnt=res_data.output_frames_gen;
            nsamp-=res_data.input_frames_used;
            samples+=res_data.input_frames_used;
            for (i=0;i<cnt;i++) *bs++=do_contrast(res_data.data_out[i]) * 32767;
        } else {
            nsamp-=cnt;
            for (i=0;i<cnt;i++) *bs++=do_contrast(*samples++) * 32767;
        }
        rc=coder_encode(buffer,cnt);
        if (rc<0) return rc;
    }
    return 0;
}

int this_part;

int encoder_init_partial(char *fname)
{
    char *c,*d;int rc;
    static char namebuf[256];
    if (set_type == 0) return 0;
    if (fname) {
        local_fname=strdup(fname);
        this_part=0;
        return 0;
    }
    if (encoder_initialized) coder_finish();
    encoder_initialized=1;
    partial_samples=0;
    strcpy(namebuf,local_fname);
    c=strrchr(local_fname,'/');
    if (c) c++;else c=local_fname;
    d=strrchr(c,'.');
    this_part++;
    if (d && d!=c) {
        sprintf(namebuf+(d-local_fname),"_%03d%s",this_part,d);
    }
    else {
        if (set_type == 1) d=".mp3";
        else if (set_type == 2) d=".wav";
        else d=".awb";
        sprintf(namebuf+strlen(namebuf),"_%03d%s",this_part,d);
    }
    rc=coder_init(namebuf);
    encoder_initialized=1;
    return rc;
}


int encoder_init(char *fname)
{
    if (set_type == 0) return 0;
    if (fname) output_is_stdout=!strcmp(fname,"-");
    if (part_time) return encoder_init_partial(fname);
    if (!fname) {
        if (encoder_initialized) return 0;
        if (!local_fname) {
            errno=1;
            return -1;
        }
        if (coder_init(local_fname)) return -1;
        encoder_initialized=1;
        return 0;
    }
    local_fname=strdup(fname);
    return 0;
}

static int put_silence_to_encoder(int count);

int encoder_finish(void)
{
    if (!encoder_initialized) return 0;
    if (resampler) {
        src_delete(resampler);
        free(res_data.data_in);
        free(res_data.data_out);
        resampler=NULL;
    }
    if (sonic) {
        sonicDestroyStream(sonic);
        sonic=NULL;
    }
    if (output_padding > 0) {
		put_silence_to_encoder((output_padding * output_samplefreq) / 1000);
	}
    encoder_initialized=0;
    return coder_finish();
}



static double compute_frame_energy(void)
{
    int i;
    double t=0;
    for (i=0;i<frame_buffer_pos;i++) t+=fabs(frame_buffer[i]);
    return t/frame_buffer_size;
}

static void feed_pause_buffer(void)
{
    if (pause_count < pause_left) {
        simple_memcpy(pause_buffer+pause_count * frame_buffer_size,
            frame_buffer,
            frame_buffer_size*sizeof(float));
    }
    else {
        int pc=pause_count-pause_left;
        if (pc >= pause_right) {
            if (pause_right > 1) {
                simple_memcpy(
                    pause_buffer_right,
                    pause_buffer_right + frame_buffer_size,
                    (pause_right-1) * frame_buffer_size*sizeof(float));
            }
        }
        simple_memcpy(pause_buffer_right+(pause_right-1) * frame_buffer_size,
                frame_buffer,
                frame_buffer_size*sizeof(float));
    }
    pause_count++;
}


int put_samples_to_encoder(float *samples,int nsamp)
{
    static float buf[4096];
    int n;


    if (!(use_feats & AUDINA_FEAT_TEMPO)) {
        return encode_samples(samples,nsamp);
    }
    if (!sonic) sonic=sonicCreateStream(tempo,samplefreq);
    sonicWriteToStream(sonic,samples,nsamp);
    for (;;) {
        n=sonicReadFromStream(sonic,buf,4096);
        if (n<=0) break;
        n=encode_samples(buf,n);
        if (n) return n;
    }
    return 0;
}

static int put_silence_to_encoder(int count)
{
    int n;
    static float buf[4096];
    static int ms=0;
    if (!ms) {
        memset(buf,0,sizeof(buf));
        ms=1;
    }
    while (count > 0) {
        n=count;
        if (n>4096) n=4096;
        count-=n;
        if (put_samples_to_encoder(buf,n)) return -1;
    }
    return 0;
}

/*


*/

static int sum_pau,cnt_pau;

static int should_break_part(int opc)
{
    int partm;
    double last_ti;
    if (!part_time) return 0;
    partm=partial_samples/output_samplefreq;
    if (sample_length <= 0) {
        if (partm > part_time) return 1;
        if (opc > 1.5 * (sum_pau/cnt_pau) && partm > 0.93 * part_time) {
            return 1;
        }
        if (opc > sum_pau/cnt_pau && partm > 0.96 * part_time) {
            return 1;
        }
        return 0;
    }
    if (current_samples >= sample_length) return 0;
    last_ti=(sample_length-current_samples)/(samplefreq*tempo);
    if (last_ti >= 2* part_time) {
        if (partm > part_time) {return 1;}
        if (opc > 1.5 * (sum_pau/cnt_pau) && partm > 0.93 * part_time) {
            return 1;
        }
        if (opc > sum_pau/cnt_pau && partm > 0.96 * part_time) {
            return 1;
        }
        return 0;
    }
    if (last_ti < part_time) return 0;
    if (partm > last_ti) {return 1;}
    if (opc > 1.5 * (sum_pau/cnt_pau) && partm > 0.92 * last_ti) {
        return 1;
    }
    if (opc > sum_pau/cnt_pau && partm > 0.96 * last_ti) {
        return 1;
    }
    return 0;
}

static int unfeed_pause_buffer(int mode)
{
    int m,n,i,paulen,pausamp,opc;
    if (!mode) { //init
        n=pause_count;
        if (n>min_pause_frames) n=min_pause_frames;
        m=pause_count-n;
        m*=frame_buffer_size;
        n*=frame_buffer_size;
        pause_count=0;
        if (n) return put_samples_to_encoder(pause_buffer+m,n);
        return 0;
    }
    if (mode < 0) { // final
        n=pause_count;
        if (n>pause_left) n=pause_left;
        pause_count=0;
        if (n) return put_samples_to_encoder(pause_buffer,n*frame_buffer_size);
        return 0;
    }
    opc=pause_count;
    if (pause_count <=min_pause_frames) {
        if (pause_count) n=put_samples_to_encoder(pause_buffer,
            frame_buffer_size * pause_count);
        else n=0;
        pause_count=0;
        return n;
    }
    opc=pause_count;
    sum_pau+=opc;
    cnt_pau++;
    paulen=(pause_count-1) * frame_buffer_size * pause_length +frame_buffer_size;
    if (paulen > max_pause_samples) paulen=max_pause_samples;
    else if (special_flags & AUDINA_F_TUNE) paulen=max_pause_samples;
    n=pause_count;
    if (n>pause_buffer_frames) n=pause_buffer_frames;
    pausamp=n * frame_buffer_size;
    pause_count=0;
    if (paulen < pausamp) {
        for (i=0;i<paulen;i++) {
            second_buffer[i]=
                (pause_buffer[i]*(paulen-i) +
                pause_buffer[pausamp-paulen+i] * i)/paulen;
        }
        if (should_break_part(opc)) {
            if (put_samples_to_encoder(second_buffer,paulen/2)) return -1;
            if (encoder_init_partial(NULL)) return -1;
            return put_samples_to_encoder(second_buffer+paulen/2,paulen/2);
        }
        return put_samples_to_encoder(second_buffer,paulen);
    }
    if (paulen == pausamp) {
        if (should_break_part(opc)) {
            if (put_samples_to_encoder(pause_buffer,paulen/2)) return -1;
            if (encoder_init_partial(NULL)) return -1;
            return put_samples_to_encoder(pause_buffer+paulen/2,paulen/2);
        }
        return put_samples_to_encoder(pause_buffer,paulen);
    }
    if (pausamp < pause_buffer_frames * frame_buffer_size) {
        n=pausamp/2;
        m=pausamp-n;
    }
    else {
        n=pause_left * frame_buffer_size;
        m=pause_right * frame_buffer_size;
    }
    for (i=0;i<n;i++) {
        pause_buffer[i]=(pause_buffer[i] * (n-i))/n;
    }
    if (put_samples_to_encoder(pause_buffer,n)) return -1;

    put_silence_to_encoder(paulen-pausamp);
    if (should_break_part(opc)) {
        if (encoder_init_partial(NULL)) return -1;
    }
    for (i=0;i<m;i++) {
        pause_buffer[i]=(pause_buffer[i+n] * i)/m;
    }
    if (put_samples_to_encoder(pause_buffer,m)) return -1;
    return 0;
}

/* przejście pauza-voice:
 * dla pause_count <= 3 po prostu flush
 * długość pauzy=
 *  (pause_count-1) * frame_buffer_size * pause_length +frame_buffer_size;
 * teraz korekta dla max_pause_samples
 * oraz min_pause_samples
 * i tyle sampli trzeba wyciepać na wyjście */

void do_equalize(float *samples,int nsamp)
{
    int i;
    for (i=0;i<nsamp;i++) samples[i]=equalise(&eq_std,samples[i])*eqgain;
}
void print_maxima(float *samples,int nsamp)
{
    static int scount=0;
    static double smax=0,dmax=0;
    int i,j;
    char sbuf[40],ubuf[32];
    if (!sample_length) strcpy(ubuf,"-");
    else sprintf(ubuf,"%5.1f",100.0*((double)current_samples/(double)sample_length));
    for (i=0;i<nsamp;i++) {
        double d=fabs(samples[i]);
        if (d > smax) smax=d;
        scount++;
        if (scount >=8192) {
            scount=0;
            if (smax>dmax) dmax=smax;
            for (j=0;j<40;j++) sbuf[j]=(j<40*smax)?'*':'-';
            fprintf(stderr,"%s %s %5.3f   \r",ubuf,sbuf,dmax);
            smax=0;
        }
    }
    if (!nsamp) fprintf(stderr,"\n");
}

int put_samples(float *samples,int nsamp)
{
    int is_sil,n,change_phase;
    float buf[4096];
    if (special_flags & AUDINA_F_VUMETER) print_maxima(samples,nsamp);
    if (samples && force_finish) {
        put_samples(NULL,0);
        exit(0);
    }
    if (!(use_feats & AUDINA_FEAT_PAUSES)) {
        if (use_equalizer & 1) do_equalize(samples,nsamp);
        return put_samples_to_encoder(samples,nsamp);
    }
    if (!nsamp) {
        if (file_phase == 2) unfeed_pause_buffer(-1);
        if (sonic) {
            sonicFlushStream(sonic);
            n=sonicReadFromStream(sonic,buf,4096);
            if (n>=0) {
                n=encode_samples(buf,n);
                if (n) return n;
            }
            sonicDestroyStream(sonic);
            sonic=NULL;
        }
        return encode_samples(NULL,0);
    }
    for (;;) {
        while (frame_buffer_pos < frame_buffer_size && nsamp > 0) {
            frame_buffer[frame_buffer_pos++]=*samples++;
            nsamp--;
        }
        if (frame_buffer_pos < frame_buffer_size) return 0;
        double t=compute_frame_energy();
        if (use_equalizer & 1) do_equalize(frame_buffer,frame_buffer_size);
        frame_buffer_pos=0;
        if (special_flags & AUDINA_F_DUMPF) fprintf(stderr,"%9.3f\n",(t * density)/SILENCE_LIMIT);
        is_sil=(t*density)<SILENCE_LIMIT;
        if (is_sil) change_phase=file_phase?2:0;
        else change_phase=1;
        if (change_phase == 0) continue;
        if (change_phase == 1) {
            if (unfeed_pause_buffer(file_phase)) return -1;
            file_phase=change_phase;
            n=put_samples_to_encoder(frame_buffer,frame_buffer_size);
        }
        else {
            feed_pause_buffer();
        }
        file_phase=change_phase;
    }
    return 0;
}

