/*
 * mpg123.c - part of Audina audiobook stretcher
 * Copyright (C) Bohdan R. Rau 2010-2011 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpg123.h>
#include "audina.h"
#include "netinet/in.h"

static mpg123_handle *mh;
static float *mh_buffer;
static int mh_bufsize;

static void m123_cleanup(mpg123_handle *mh)
{
    mpg123_close(mh);
    mpg123_delete(mh);
    mpg123_exit();
    if (mh_buffer) free(mh_buffer);
    mh_buffer=NULL;
}

static void m123_error(int err)
{
    fprintf(stderr,"mpg123 problem: %s\n",mh?mpg123_strerror(mh):mpg123_plain_strerror(err));
    m123_cleanup(mh);
}

static mpg123_id3v1 *idv1;
static mpg123_id3v2 *idv2;

static void m123_id3(void)
{
    int err=mpg123_id3(mh,&idv1,&idv2);
    if (err != MPG123_OK) {
        fprintf(stderr,"ID3 problem: %s\n",mpg123_strerror(mh));
        return;
    }
    if (idv2) {
        if (!title && idv2->title && idv2->title->fill > 1 && *idv2->title->p) title=id3_strdup(idv2->title->p);
        if (!album && idv2->album && idv2->album->fill > 1 && *idv2->album->p) album=id3_strdup(idv2->album->p);
        if (!artist && idv2->artist && idv2->artist->fill > 1 && *idv2->artist->p) artist=id3_strdup(idv2->artist->p);
    }
    if (idv1) {
        if (!title && idv1->title[0]) title=id3_strdup(idv1->title);
        if (!album && idv1->album[0]) album=id3_strdup(idv1->album);
        if (!artist && idv1->artist[0]) artist=id3_strdup(idv1->artist);
    }
}

int mp3_is_wrap(char *filename)
{
    FILE *f;char buf[6];
    int rc=0;
    int size;
    f=fopen(filename,"r");
    if (!f) return 0;
    if (!fread(buf,6,1,f)) goto eout;
    if (strncmp(buf,"ID3",3)) {
        fseek(f,0,SEEK_SET);
        goto readwrap;
    }
    if (!fread(&buf,4,1,f)) goto eout;
    size=(buf[0] << 21) | (buf[1] << 14) | (buf[2] << 7) | buf[3];
    fseek(f,size,SEEK_CUR);
readwrap:
    if (!fread(&buf,4,1,f)) goto eout;
    if (!strncmp(buf,"WRAP",4)) rc=1;
eout:
    fclose(f);
    return rc;
}


int input_convert_mpg123(char *filename)
{
    int encoding=0,channels=0,rc;
    long rate=0;
    if (mp3_is_wrap(filename)) {
        fprintf(stderr,"File is mp3wrap archive. Use mp3splt tool to extract tracks\nor turn on ffmpeg decoder to force decoding as plain mp3 file.\n");
        return -1;
    }
    int err = mpg123_init();

    if( err != MPG123_OK || (mh = mpg123_new(NULL, &err)) == NULL
            
            || mpg123_param(mh,MPG123_ADD_FLAGS,MPG123_MONO_MIX | MPG123_FORCE_FLOAT,0.0) != MPG123_OK
            
            /* Let mpg123 work with the file, that excludes MPG123_NEED_MORE messages. */
            || mpg123_open(mh, filename) != MPG123_OK
            
            /* Peek into track and get first output format. */
            || mpg123_getformat(mh, &rate, &channels, &encoding) != MPG123_OK ) {
                m123_error(err);
                return -1;
    }
    mpg123_format_none(mh);
    encoding=MPG123_ENC_FLOAT_32;
    err=mpg123_format(mh, rate, 1, encoding);
    if (err != MPG123_OK) {
        m123_error(err);
        return -1;
    }
    err=mpg123_getformat(mh, &rate, &channels, &encoding);
    if (err != MPG123_OK) {
        m123_error(err);
        return -1;
    }
    if (encoding != MPG123_ENC_FLOAT_32 || channels != 1) {
        fprintf(stderr,"mpg123: cannot set mono 32-bit float encoding\n");
        m123_cleanup(mh);
        return -1;
    }
    mh_bufsize=4*mpg123_outblock(mh);
    mh_buffer=malloc(mh_bufsize);
    sample_length=mpg123_length(mh);
    if (sample_length == MPG123_ERR) {
        fprintf(stderr,"Cannot find sample length: %s\n",mpg123_strerror(mh));
        sample_length=0;
    }
    m123_id3();
    if (skip_time) {
        off_t sampleoffset=rate * (off_t) skip_time;
        off_t sampleoff=mpg123_seek(mh,sampleoffset,SEEK_SET);
        if (sampleoff == MPG123_ERR) {
            fprintf(stderr,"Cannot seek: %s\n",mpg123_strerror(mh));
            m123_cleanup(mh);
            return -1;
        }
        if (sampleoff != sampleoffset) {
            fprintf(stderr,"Cannot seek for unknown reason, try \"-i f\" option\n");
            m123_cleanup(mh);
            return -1;
        }
        skip_time=sampleoffset/rate;
        if (sample_length>0) sample_length-=sampleoffset;
    }
    rc=set_input_samplerate(rate);
    if (rc<0) {
        m123_cleanup(mh);
        return rc;
    }
    current_samples=0;
    for (;;) {
        size_t done;
        err=mpg123_read(mh,(void *)mh_buffer,mh_bufsize,&done);
        if (err == MPG123_OK) {
            int i;
            done = done/4;
            current_samples+=done;
            for (i=0;i<done;i++) mh_buffer[i]=mh_buffer[i]*mad_gain;
            int rc=put_samples(mh_buffer,done);
            if (rc<0) {
                m123_cleanup(mh);
                return rc;
            }
            continue;
        }
        if (err != MPG123_DONE) {
            m123_error(err);
            return -1;
        }
        break;
    }
    rc=put_samples(NULL,0);
    m123_cleanup(mh);
    return rc;
}
